#ifndef _SONY_FPGA_CMD_H_
#define _SONY_FPGA_CMD_H_
/**
 * @file sony_fpga_cmd.h
 * @brief sony_fpga外部インタフェース定義ファイル
 */
#include <linux/ioctl.h>
#include <linux/types.h>

#define SONY_FPGA_IOC_MAGIC                (0x73)

/**
 * @brief VideoDescriptor設定シーケンス<br>
 * ライフラインについては抽象的に表記しています。(FPGAとドライバを一緒くたに表記する等)
 *
 * @startuml
 * title Video Dscriptor設定シーケンス
 * actor User
 * create OS
 * [->OS : 起動
 *
 * create "gif-video"
 * create "gif-fpga"
 * create GPU
 * [->GPU : 起動
 * OS -> "gif-video" : Probe
 * activate "gif-video"
 * note right of "gif-video"
 *   Descriptor領域の生成、Video DMA Portの設定等を行う.
 * end note
 * return
 *
 * OS -> "gif-fpga" : Probe
 * activate "gif-fpga"
 * return
 *
 * create User
 * activate User
 * OS -> User : 起動
 *
 * group Video関連初期設定
 *   User -> GPU : GPUメモリ確保
 *   activate GPU
 *   |||
 *   return GPUメモリ
 *   User -> GPU : Attribute Page固定設定 (PIN)
 *   activate GPU
 *   |||
 *   return
 *   |||
 *   User -> "gif-video" : SONY_FPGA_PIN_GPU_MEMORY (GPUメモリ)
 *   activate "gif-video"
 *   "gif-video" -> GPU : nvidia_p2p_get_pages
 *   activate GPU
 *   note right of "gif-video"
 *     物理ページ情報を取得することでDMACに通知できるようにする
 *   end note
 *   return
 *   return
 *
 *   ||50||
 *
 *   loop 0 , Descriptorの数
 *     |||
 *     note over User, GPU
 *       Descriptorは各Port(9port) x ch(1-4ch) x R/W x strip(0-3 or 2?) 個作成する
 *       4Kフォーマットの場合は1ch目だけ設定する。
 *     end note
 *     |||
 *     User -> "gif-video" : SONY_FPGA_MAKE_VIDEO_DESCRIPTOR(gpuオフセット,size)
 *     activate "gif-video"
 *     return handle (Descriptorを識別するもの)
 *     |||
 *     User -> "gif-video" : SONY_FPGA_SET_VIDEO_DESCRIPTOR(port, handle, ch, strip, direction)
 *     activate "gif-video"
 *     note over User, GPU
 *       各DMAポート毎のdmac_mediaレジスタを設定する。
 *     end note
 *     return
 *     |||
 *   end
 *
 *   ||50||
 *
 *   User -> "gif-video" : SONY_FPGA_SET_VIDEO_DMAC(1: enable)
 *   activate "gif-video"
 *   note over User , GPU
 *     Video DMAを開始する。※ 開始タイミングについてはFPGA制約がないか確認が必要
 *   end note
 *   return
 *   |||
 * end
 *
 * == 初期化完了 ==
 *
 * User -> "gif-fpga" : SONY_FPGA_WAIT_VD_INFO
 * activate "gif-fpga"
 *
 * == INT_COUNT(0) ==
 * [->"gif-video" : VD割り込み(0)
 * activate "gif-video"
 * "gif-video"->"gif-video" : Video DMA Read(strip 0)
 * "gif-video"->"gif-video" : Video DMA Write(strip 0)
 * return int_count(0)
 * deactivate "gif-video"
 *
 * note over User , GPU
 * GPU演算処理をstrip毎に行う。
 * end note
 *
 * User -> "gif-fpga" : SONY_FPGA_WAIT_VD_INFO
 * activate "gif-fpga"
 *
 * == INT_COUNT(1) ==
 * [->"gif-video" : VD割り込み(1)
 * activate "gif-video"
 * "gif-video"->"gif-video" : Video DMA Read(strip 1)
 * "gif-video"->"gif-video" : Video DMA Write(strip 1)
 * return int_count(1)
 * deactivate "gif-video"
 *
 * User -> "gif-fpga" : SONY_FPGA_WAIT_VD_INFO
 * activate "gif-fpga"
 *
 * == INT_COUNT(2) ==
 * [->"gif-video" : VD割り込み(2)
 * activate "gif-video"
 * "gif-video"->"gif-video" : Video DMA Read(strip 2)
 * "gif-video"->"gif-video" : Video DMA Write(strip 2)
 * return int_count(2)
 * deactivate "gif-video"
 * hide footbox
 * @enduml
 * ---
 */

/**
 * @brief VD同期レジスタR/Wシーケンス
 * @startuml
 * title VD同期レジスタR/Wシーケンス
 * actor SelfCtrl
 * == サービス起動 ==
 *
 * create SelfCtrl
 * create InitScript
 * SelfCtrl -> InitScript : 実行
 * activate InitScript
 * create gif
 * InitScript -> gif: 生成
 * create swr
 * InitScript -> swr : 生成
 * create sdi
 * InitScript -> sdi : 生成
 * note over InitScript , sdi
 *   ドライバ初期化処理(制御DMAC初期設定)
 * end note
 * return
 * deactivate InitScript
 *
 * create User
 * SelfCtrl -> User : 起動
 * User -> gif : mmap
 * activate gif
 * return DMAバッファへのポインタ
 * User -> swr : mmap
 * activate swr
 * return DMAバッファへのポインタ
 * User -> sdi : mmap
 * activate sdi
 * return DMAバッファへのポインタ
 * note over gif , User
 *   Write DMAはFPGAからCPUへのWriteを指す。
 *   Userから見るとFPGAから情報をReadすることに相当する。
 *   mmapで返るDMAバッファ領域は
 *   Write DMA Buffer (3Page) _ Read DMA Buffer (任意サイズ)となっており、
 *   Read DMA Bufferについてはデバッグ用途でのみ使用する。
 *   Write DMA Bufferは2Page分確保しているが、
 *   FPGAが書き込むデータはFPGA毎に異なる。(SDI Informationの有無等)
 * end note
 *
 * ... アプリケーション初期化処理等...
 *
 * note over User
 *   定常状態へ遷移
 * end note
 *
 * User -> gif : SONY_FPGA_WAIT_VD_INFO
 * activate gif
 * note over gif , User
 *   Write DMAC (ステータスのRead)はint_count(0,4,8...のタイミングで自動的に動作する)
 * end note
 *
 * [->gif : Write DMAC完了(Diag Status 0, Diag Staus 1)
 * [->swr : Write DMAC完了(Diag Status 0, Diag Staus 1)
 * [->sdi : Write DMAC完了(Diag Status 0, Diag Staus 1, SDI Information, Audio/Meta)
 * |||
 * [->gif : VD割り込み(0)
 * return int_count (0)
 *
 * == int_count(0) ==
 * User -> gif : SONY_FPGA_REG_WRITE(Resizer(CH1))
 * activate gif
 * |||
 * return
 *
 * User -> gif : SONY_FPGA_WAIT_VD_INFO
 * activate gif
 * [->gif : VD割り込み(1)
 * return int_count (1)
 *
 * == int_count(1) ==
 * User -> gif : SONY_FPGA_REG_WRITE(MultiViewer(CH1))
 * activate gif
 * |||
 * return
 * User -> gif : SONY_FPGA_REG_WRITE(REF_MGR)
 * activate gif
 * |||
 * return
 *
 * User -> gif : SONY_FPGA_WAIT_VD_INFO
 * activate gif
 * [->gif : VD割り込み(2)
 * return int_count (2)
 *
 * == int_count(2) ==
 * note over gif , User
 *   SONY_FPGA_REG_WRITEは機能ブロック単位に呼び出すこと。タイミングや対象は一例として記載している。
 * end note
 * User -> swr : SONY_FPGA_REG_WRITE(MIX(PP))
 * activate swr
 * |||
 * return
 * User -> swr : SONY_FPGA_REG_WRITE(MIX(ME))
 * activate swr
 * |||
 * return
 * User -> sdi : SONY_FPGA_REG_WRITE(FC)
 * activate sdi
 * |||
 * return
 * User -> sdi : SONY_FPGA_REG_WRITE(HDR)
 * activate sdi
 * |||
 * return
 * User -> gif : SONY_FPGA_WAIT_VD_INFO
 * activate gif
 * [->gif : VD割り込み(3)
 * return int_count (3)
 *
 * == int_count(3) ==
 * alt レジスタ設定完了 case
 *     note over gif , User
 *       必ずしもint_count(3)である必要はない。
 *       各FPGAを同じint_countで設定する必要はない。
 *     end note
 *     User -> gif : SONY_FPGA_REG_UPDATE_OK
 *     activate gif
 *     return
 *     User -> swr : SONY_FPGA_REG_UPDATE_OK
 *     activate swr
 *     return
 *     User -> sdi : SONY_FPGA_REG_UPDATE_OK
 *     activate sdi
 *     return
 * end
 * |||
 * alt 全FPGAレジスタ設定完了 case
 *     note over gif , User
 *       全FPGAのSONY_FPGA_REG_UPDATE_OK呼び出し後に呼び出すこと。
 *       必ずしもint_count(3)である必要はない。
 *     end note
 *     User -> gif : SONY_FPGA_REG_UPDATE_COMPLETED (更新領域の更新)
 *     activate gif
 *     return
 * end
 *
 * par
 * note over gif , User
 *   15msで自動起動。1ms以内に完了することをH/Wにて保証する。
 *   更新判定領域が更新されていない場合は、FPGAへの書込みは行われない。
 * end note
 * gif -> gif : Read DMA + FGPAレジスタ更新
 * else
 * swr -> swr : Read DMA + FPGAレジスタ更新
 * else
 * sdi -> sdi : Read DMA + FPGAレジスタ更新
 * end
 * [->gif : VD割り込み(4)
 * ||100||
 * hide footbox
 * @enduml
 */

/**
 * ---
 * @brief @b デバイスファイル
 * @details     /dev/sony_fpga_xxという名称でデバイスファイルが生成される。xxはpcieのbusnoを示す。
 */

/**
 * ---
 * @brief @b ファイルオペレーション
 *
 *  ---
 *
 * @details @b mmap <br>
 *         制御DMAC用のDMAバッファを仮想アドレス空間にマッピングする。
 * @return 仮想アドレス
 * @sa mmap
 * @details 使用方法はPOSIX mmapのとおり。<br>
 * 先頭2PAGE分がWrite DMA Buffer領域、それ以降がRead DMA Buffer領域となる。Read DMA BufferサイズはFPGA毎に異なる。
 *
 * @note
 *  制御DMAC搭載FPGAでのみ有効<br>
 *  Write DMA とはDMACがWriteするという意味なのでFPGAのステータス等をリードすることを示す。<br>
 *  Read DMA とはDMACがReadするという意味なので、アプリケーションからの設定データをDMAで読み、FPGAに反映する。
 *
 *  ---
 */

/**
 * @name 外部インタフェース
 *   @b ioctl <br>
 *   有効なIOCTL番号定義を記述する
 * @{
 */

#ifndef GPU_DISABLE
/**
 * @brief GPU Memory PIN設定
 * @param[in] struct gpu_mem_info
 * @note GIF-Videoでのみ有効 <br>
 *  GPU非搭載時は未定義
 */
#define SONY_FPGA_PIN_GPU_MEMORY      _IOW(SONY_FPGA_IOC_MAGIC, 1, struct gpu_mem_info)
#endif

/**
 * @brief VD割り込み受信
 * @param[out] u32 割り込み位置番号 0から23までのカウンター値
 * @note GIF-FPGAでのみ有効
 */
#define SONY_FPGA_WAIT_VD_INFO    _IOR(SONY_FPGA_IOC_MAGIC, 2, __u32)


/**
 * @brief VD割り込み設定
 * @param[in] u32 0: Disable, 1: Enable(Default)
 */
#define SONY_FPGA_VD_INT_ENABLE   _IOW(SONY_FPGA_IOC_MAGIC, 3, __u32)

/**
 * @brief レジスタ書込み設定
 * @param[in] struct fpga_reg_write_arg
 *
 * @note 機能ブロック毎に設定すること。設定する順序は順不同で良い。
 * @sa SONY_FPGA_CTRL_DMAC_IGNORE_ADDRESS <br>
 *  module_param_ctrl_dmac_debug_enable = 1のとき無視リストに登録されているベースアドレスについては書込みを抑止する。
 */
#define SONY_FPGA_REG_WRITE       _IOW(SONY_FPGA_IOC_MAGIC, 4, struct fpga_reg_write_arg)

/**
 * @brief 全FPGAレジスタ更新完了通知
 * @note 全てのレジスタ更新設定が完了後に1VD内で1度だけ呼ぶこと。GIF-FPGAでのみ有効
 */
#define SONY_FPGA_REG_UPDATE_COMPLETED   _IO(SONY_FPGA_IOC_MAGIC, 5)

/**
 * @brief レジスタ更新完了通知
 * @note 各FPGAのレジスタ更新が完了したことを通知する
 */
#define SONY_FPGA_REG_UPDATE_OK   _IO(SONY_FPGA_IOC_MAGIC, 6)

#ifndef GPU_DISABLE
/**
 * @brief Video転送用 Descriptor作成
 * @param[in] struct video_dmac_region
 * @sa SONY_FPGA_PIN_GPU_MEMORY
 * @note GPU非搭載時は未定義
 */
#define SONY_FPGA_MAKE_VIDEO_DESCRIPTOR  _IOWR(SONY_FPGA_IOC_MAGIC, 7, struct video_dmac_region)

/**
 * @brief Video転送用 Descriptor領域初期化
 * @param なし
 * @details SONY_FPGA_MAKE_VIDEO_DESCRIPTORで設定した情報を消去する<br>
 *  以降はvideo_descriptor領域の先頭から割り当てる。
 * @note debug用途。本番では使用しない予定 <br>
 *  GPU非搭載時は未定義
 */
#define SONY_FPGA_RESET_VIDEO_DESCRIPTOR  _IO(SONY_FPGA_IOC_MAGIC, 8)

/**
 * @brief Video転送用 Descriptor設定
 * @param[in] struct video_dmac_descriptor
 * @note GPU非搭載時は未定義
 */
#define SONY_FPGA_SET_VIDEO_DESCRIPTOR  _IOW(SONY_FPGA_IOC_MAGIC, 9, struct video_dmac_descriptor)

/**
 * @brief Video転送用 Descriptor設定
 * @param[in] u32 enable ::SONY_FPGA_ENABLE , ::SONY_FPGA_DISABLE
 * @note GIF-Videoでのみ有効 <br>
 *  GPU非搭載時は未定義
 */
#define SONY_FPGA_SET_VIDEO_DMAC  _IOW(SONY_FPGA_IOC_MAGIC, 10, __u32)

#endif

/**
 * @brief Master Mode設定
 * @param[in] u32 master_mode @n
 *                [7:4]: Slave FPGA数 @n
 *                [0]: 0: Slave, 1: Master
 * @note for Debug
 */
#define SONY_FPGA_SET_MASTER_MODE _IOW(SONY_FPGA_IOC_MAGIC, 11, __u32)

/**
 * @brief Ctrl DMAC設定
 * @param[in] u32 enable ::SONY_FPGA_ENABLE , ::SONY_FPGA_DISABLE
 */
#define SONY_FPGA_SET_CTRL_DMAC   _IOW(SONY_FPGA_IOC_MAGIC, 12, __u32)

/**
 * @brief Multicast設定
 * @param[in] u32 enable ::SONY_FPGA_ENABLE , ::SONY_FPGA_DISABLE
 * @note GIF-FPGAでのみ有効
 */
#define SONY_FPGA_SET_MULTICAST   _IOW(SONY_FPGA_IOC_MAGIC, 13, __u32)

/**
 * @brief Multicast Address設定
 * @param[in] struct multicast_info
 * @note GIF-FPGAでのみ有効
 */
#define SONY_FPGA_SET_MULTICAST_ADDRESS   _IOW(SONY_FPGA_IOC_MAGIC, 14, struct multicast_info)


/**
 * @brief 制御DMAC無視リスト設定
 * @param[in] struct ctrl_dmac_ignore_list
 * @details Read DMAバッファに登録する際に無視するベースアドレスを登録する。<br>
 * 任意のFPGAで実行した結果が全FPGAに反映される。
 * @note デバッグ用途
 * @sa SONY_FPGA_REG_WRITE
 */
#define SONY_FPGA_CTRL_DMAC_IGNORE_ADDRESS _IOW(SONY_FPGA_IOC_MAGIC, 15, struct ctrl_dmac_ignore_list)

/**
 * @brief 制御DMAC無視リストクリア
 * @details SONY_FPGA_CTRL_DMAC_IGNORE_ADDRESSで設定した無視リストをクリアする。<br>任意のFPGAで実行した結果が全FPGAに反映される。
 * @note デバッグ用途。 排他制御等は未実装。
 */
#define SONY_FPGA_CTRL_DMAC_IGNORE_LIST_CLEAR _IO(SONY_FPGA_IOC_MAGIC, 16)

/**
 * @brief DMAC_FILE ステータス取得
 * @details dmac_file空間のmodule_statusを返す
 */
#define SONY_FPGA_GET_DMAC_FILE_STATUS _IOR(SONY_FPGA_IOC_MAGIC, 17, __u32)

/**
 * @brief Video File Write
 * @details Video FileをDDRに書き込む
 * @param[in] struct video_file_info
 * @return  0: 正常終了 , 0以外: 異常終了 (errnoを参照)
 * @note ドライバ内で、DMAの完了を待機する。<br>
 *       timeout時間は1sec(暫定)とする。<br>
 *       DMA実行中に実行されたら即時エラーを返す。(前優先) <br>
 *       GIF FPGA以外で実行された場合はなにもしない。
 */
#define SONY_FPGA_VIDEO_FILE_WRITE _IOWR(SONY_FPGA_IOC_MAGIC, 18, struct video_file_info)

/**
 * @brief Video File Read
 * @details Video FileをDDRから読み出す
 * @param[in] struct video_file_info
 * @return  0: 正常終了 , 0以外: 異常終了 (errnoを参照)
 * @note ドライバ内で、DMAの完了を待機する。<br>
 *       timeout時間は1sec(暫定)とする。<br>
 *       DMA実行中に実行されたら即時エラーを返す。(前優先) <br>
 *       GIF FPGA以外で実行された場合はなにもしない。
 */
#define SONY_FPGA_VIDEO_FILE_READ _IOWR(SONY_FPGA_IOC_MAGIC, 19, struct video_file_info)

/**
 * @brief SDI OUT CH Enable設定
 * @details SDI OUT CH Enable/Disableを設定する
 * @param[in] __u32 : [31:24] = 0h , [23:00] = SDI OUT 24ch 〜 1chのEnable(0)/Disable(1) ビット 出力抑止対象のCHのBITを立てる
 * @return  0: 正常終了 , 0以外: 異常終了 (errnoを参照)
 * @note ドライバ内で、NIOSステータス監視を行うため長時間のブロックの可能性あり。(最大90ms程度)<br>
 *       SDI FPGA専用のため、SDI FPGA以外で実行された場合は無処理で正常終了を返す。
 */
#define SONY_FPGA_SET_SDI_OUT_CH               _IOW(SONY_FPGA_IOC_MAGIC, 20, __u32)

/**
 * @brief Ref Lock通知
 * @details Ref LockしたことをSlave基板に通知し、VideoFormat処理を実行させる
 * @param[in] struct video_file_info
 * @return  0: 正常終了 , 0以外: 異常終了 (errnoを参照)
 * @note ドライバ内で、DMAの完了を待機する。<br>
 *       timeout時間は1sec(暫定)とする。<br>
 *       DMA実行中に実行されたら即時エラーを返す。(前優先)
 */
#define SONY_FPGA_REF_LOCK                     _IO(SONY_FPGA_IOC_MAGIC, 21)

/**
 * @brief mute設定
 * @details Audioのmuteを設定する
 * @param[in] __u32 mute<br>
 *     mute[09:00] : CH1〜10を示す(LSB CH1) <br>
 *     mute[25:16] : CH11〜20を示す(LSB CH11) <br>
 *     各CH毎に 0x1: Mute On / 0x0: Mute Offを設定する
 *
 * @return  0: 正常終了 , 0以外: 異常終了 (errnoを参照)
 * @note SDI FPGA以外で実行された場合はなにもせず正常終了する。
 */
#define SONY_FPGA_AUDIO_MUTE                   _IOW(SONY_FPGA_IOC_MAGIC, 22, __u32)

/**
 * @brief Audio File Write
 * @details Audio FileをHBM2に書き込む
 * @param[in] struct audio_file_info
 * @return  0: 正常終了 , 0以外: 異常終了 (errnoを参照)
 * @note ドライバ内で、DMAの完了を待機する。<br>
 *       timeout時間は1sec(暫定)とする。<br>
 *       DMA実行中に実行されたら即時エラーを返す。(前優先) <br>
 *       SDI FPGA以外で実行された場合はなにもしない。
 */
#define SONY_FPGA_AUDIO_FILE_WRITE _IOWR(SONY_FPGA_IOC_MAGIC, 23, struct audio_file_info)

/**
 * @brief Audio File Read
 * @details Audio FileをDDRから読み出す
 * @param[in] struct audio_file_info
 * @return  0: 正常終了 , 0以外: 異常終了 (errnoを参照)
 * @note ドライバ内で、DMAの完了を待機する。<br>
 *       timeout時間は1sec(暫定)とする。<br>
 *       DMA実行中に実行されたら即時エラーを返す。(前優先) <br>
 *       SDI FPGA以外で実行された場合はなにもしない。
 */
#define SONY_FPGA_AUDIO_FILE_READ _IOWR(SONY_FPGA_IOC_MAGIC, 24, struct audio_file_info)

/**
 * @brief SlaveData File Write
 * @details Slave基板のアップデート用データを1ブロック転送する
 * @param[in] struct slavedata_write_info
 * @return  0: 正常終了 , 0以外: 異常終了 (errnoを参照)
 * @note ドライバ内で、DMAの完了を待機する。<br>
 *       timeout時間は1sec(暫定)とする。<br>
 *       DMA実行中に実行されたら即時エラーを返す。(前優先) <br>
 *       GIF/SDI FPGA以外で実行された場合はなにもしない。
 */
#define SONY_FPGA_SLAVEDATA_FILE_WRITE _IOWR(SONY_FPGA_IOC_MAGIC, 25, struct slavedata_write_info)

/**
 * @brief SlaveData Update Progress
 * @details Slave基板のアップデート状態を返す
 * @retval  2: Processing/Wait
 * @retval  4: Done 1Block
 * @retval  8: Error 1Block
 * @retval  1: Done All Block
 * @note ドライバ内で、DMAの完了を待機する。<br>
 *       timeout時間は1sec(暫定)とする。<br>
 *       DMA実行中に実行されたら即時エラーを返す。(前優先) <br>
 *       GIF/SDI FPGA以外で実行された場合はなにもしない。
 */
#define SONY_FPGA_SLAVEDATA_UPDATE_PROGRESS _IOR(SONY_FPGA_IOC_MAGIC, 26, __u32)

/**
 * @brief SlaveData Update Start
 * @details Slave基板のアップデートを開始を通知する
 * @param[in] struct slavedata_start_param
 * @return  0: 正常終了 , 0以外: 異常終了 (errnoを参照)
 * @note アップデート完了通知が発行されるまでPUG FILE機能の使用を制限する
 */
#define SONY_FPGA_SLAVEDATA_UPDATE_START _IOW(SONY_FPGA_IOC_MAGIC, 27, struct slavedata_start_param)

/**
 * @brief SlaveData Update Finish
 * @details Slave基板のアップデートを完了を通知する
 * @param[in] u32 終了コード
 * @return  0: 正常終了 , 0以外: 異常終了 (errnoを参照)
 * @note PUG FILEの使用制限を解除する
 */
#define SONY_FPGA_SLAVEDATA_UPDATE_FINISH _IOR(SONY_FPGA_IOC_MAGIC, 28, __u32)

/**
 * @brief SlaveData Update Error Clear
 * @details Slave基板の Error Status を Clear する
 * @param[in] u32 終了コード(Clear 前の file_target の値)
 * @return  0: 正常終了 , 0以外: 異常終了 (errnoを参照)
 * @note SONY_FPGA_SLAVEDATA_UPDATE_PROGRESS で3回エラーとなった時に Error Status を Clear するために呼ぶ
 */
#define SONY_FPGA_SLAVEDATA_UPDATE_ERROR_CLEAR _IOR(SONY_FPGA_IOC_MAGIC, 35, __u32)

/**
 * @}
 */

#define SONY_FPGA_WAIT_VD         _IO(SONY_FPGA_IOC_MAGIC, 29)

#define SONY_FPGA_READ_FRAME_CNT               _IOR(SONY_FPGA_IOC_MAGIC, 30, __u32)
#define SONY_FPGA_CMP_TEST_CTRL                _IOW(SONY_FPGA_IOC_MAGIC, 31, __u32)
#define SONY_FPGA_REFSYNC_MON_CLEAR            _IO(SONY_FPGA_IOC_MAGIC, 32)
#define SONY_FPGA_SET_VIDEO_FORMAT             _IOW(SONY_FPGA_IOC_MAGIC, 33, __u32)
#define SONY_FPGA_SET_VIDEO_TRANSFER           _IOW(SONY_FPGA_IOC_MAGIC, 34, __u32)
#define SONY_FPGA_PAUSE_VIDEO_DMAC             _IO(SONY_FPGA_IOC_MAGIC, 35)
#define SONY_FPGA_RESTART_VIDEO_DMAC           _IO(SONY_FPGA_IOC_MAGIC, 36)
#define SONY_FPGA_SET_VIDEO_DMAC_CHANNEL       _IOW(SONY_FPGA_IOC_MAGIC, 37, struct video_dmac_channel_setting)

#ifndef __KERNEL__

#include <sys/ioctl.h>

#endif

#ifndef GPU_DISABLE
struct video_dmac_region {
    uint32_t gpu_offset; //!< PINしたGPUメモリ先頭からのオフセット
    uint32_t size;       //!< サイズ PAGE_SIZE(4096byte)の倍数を設定すること
    uint32_t handle;     //!< [out] ハンドル
};

struct video_dmac_descriptor {
    uint32_t port;       //!< ポート
    uint32_t handle;     //!< ハンドル
    uint32_t ch;         //!< チャンネル
    uint32_t strip;      //!< 短冊番号
    uint32_t direction;  //!< 方向 ::DMA_DIRECTION_READ (fpga->gpu), ::DMA_DIRECTION_WRITE (gpu->fpga)
    uint32_t kind;       //!< Descriptorの種類(未定)
};

struct video_dmac_channel_setting {
    uint32_t ch;           //!< チャンネル
    uint32_t write_port;   //!< write_port setting (Enable/Disable)
    uint32_t read_port;    //!< read_port setting (Enable/Disable)
};
#endif

enum {
    SONY_FPGA_DISABLE = 0,  //!< 0:Disable
    SONY_FPGA_ENABLE        //!< 1:Enable
};

enum {
    DMA_DIRECTION_READ = 0, //!< 0:Read  (FPGA->GPU)
    DMA_DIRECTION_WRITE     //!< 1:Write (GPU->FPGA)
};

enum {
    VIDEO_FILE_FORMAT_4096 = 0, //!< Video Fileフォーマット 4K 4096
    VIDEO_FILE_FORMAT_3840,     //!< Video Fileフォーマット 4K 3840
    VIDEO_FILE_FORMAT_1920,     //!< Video Fileフォーマット HD 1920
    VIDEO_FILE_FORMAT_1280,     //!< Video Fileフォーマット HD 1280
    VIDEO_FILE_RES1,
    VIDEO_FILE_RES2,
    VIDEO_FILE_RES3,
    VIDEO_FILE_RES7,
    VIDEO_FILE_THUMB_256,       //!<Video Fileフォーマット サムネイル 256
    VIDEO_FILE_THUMB_240,       //!<Video Fileフォーマット サムネイル 240
    VIDEO_FILE_MAX
};

#ifndef GPU_DISABLE
/**
 * @brief GPUメモリ情報
 */
struct gpu_mem_info {
    void* address; //!< 仮想アドレス
    uint32_t size; //!< サイズ
};
#endif

/**
 * @brief 制御系 DMACレジスタ書込み設定
 */
struct fpga_reg_write_arg {
    uint32_t dst_address;  //!< FPGAベースアドレス
    uint32_t* reg_top;     //!< 読み出し先アドレス
    uint32_t size;         //!< サイズ
};

/**
 * @brief マルチキャストアドレス情報
 */
struct multicast_info {
    uint32_t number;
    uint64_t address;
};

/**
 * @brief 制御DMAC無視リスト
 */
struct ctrl_dmac_ignore_list {
    uint32_t* list_addr;
    uint32_t  size;
};

/**
 * @brief Video File情報 */
struct video_file_info {
    uint32_t  format;           //!< ::VIDEO_FILE_FORMAT_4096 etc
    uint64_t  dst_address;      //!< 転送先アドレス
    uint32_t* buffer_address;   //!< バッファアドレス(4096バイト境界に配置されていること)
    uint32_t  buffer_size;      //!< バッファサイズ
};

/**
 * @brief Audio Header情報
 */
struct audio_header {
    uint8_t   sample_status;  //!< 5Frame Sample Status 0:Sample, 1:Sample+1
    uint8_t   endian;         //!< Endian 0:Little, 1:Big
    uint16_t  ch_assign;      //!< Audio Channel Assign
    uint32_t  payload_size;   //!< Audio Sample Size + padding (byte)
    uint32_t  sample_size;    //!< 有効CH数 x sample数 x 3Byte(24bit)
};
typedef struct audio_header audio_header;

/**
 * @brief Audio File情報
 */
struct audio_file_info {
    audio_header head; //!< Audioヘッダ情報
    uint64_t  target_addr;    //!< HMB2内の対象アドレス(29bit : MSB3bitは0固定)
    uint32_t  next_read;      //!< Next Frame Memory Read Address
    uint32_t* buffer_address; //!< バッファアドレス(4096バイト境界に配置されていること)
    uint32_t  buffer_size;    //!< バッファサイズ(Audioデータブロックサイズ:48KB固定値)
};

/**
 * @brief SlaveData File情報
 */
struct slavedata_write_info {
    uint32_t  block_num;      //!< これから転送するブロック番号
    uint32_t* buffer_address; //!< バッファアドレス(4096バイト境界に配置されていること)
};

enum {
    INSTALL_STATUS_INIT = 0,
    INSTALL_STATUS_FINISH = 1,
    INSTALL_STATUS_PROCESSING = 2,
    INSTALL_STATUS_BLOCK_DONE = 4,
    INSTALL_STATUS_ERROR = 8,
};
enum {
    UPDATE_TARGET_CADEC, //!< CADEC
    UPDATE_TARGET_NIOS,  //!< NIOS
    UPDATE_TARGET_FPGA1, //!< FPGA1
    UPDATE_TARGET_FPGA2  //!< FPGA2
};

/**
 * @brief SlaveUpdate起動パラメタ
 */
struct slavedata_start_param {
    uint32_t  file_target;    //!< 更新対象を選択
    uint32_t  block_count;    //!< ブロック数
};

#define REGCLR_STATUS_IRQ_NIOS 1
#define REGCLR_STATUS_FILE_TARGET 2
#define REGCLR_STATUS_FILE_BLOCK 4
#define REGCLR_STATUS_CHECKSUM 8
#endif /* _SONY_FPGA_CMD_H_ */
